package com.agiledeveloper.game;

import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.stream.Stream;

import static com.agiledeveloper.game.CellState.DEAD;
import static java.util.Comparator.comparing;
import static java.util.stream.Collectors.*;

public class GameOfLife {
  public static CellState nextState(CellState currentState, int liveNeighbors) {
    return liveNeighbors == 3 || (liveNeighbors == 2 && currentState == CellState.ALIVE) ? CellState.ALIVE : DEAD;
  }

  public static Stream<Cell> generateSignalsForOneCell(Cell cell) {
    int x = cell.x();
    int y = cell.y();

    return Stream.of(new Cell(x - 1, y - 1), new Cell(x - 1, y), new Cell(x - 1, y + 1),
                   new Cell(x, y - 1), new Cell(x, y + 1),
                   new Cell(x + 1, y - 1), new Cell(x + 1, y), new Cell(x + 1, y + 1));
  }

  public static List<Cell> generateSignalsForCells(Set<Cell> cells) {
    return cells.stream()
      .sorted(comparing(Cell::x).thenComparing(Cell::y))
      .flatMap(GameOfLife::generateSignalsForOneCell)
      .toList();
  }

  public static Map<Cell, Integer> countSignals(List<Cell> signals) {
    return signals.stream()
      .collect(groupingBy(cell -> cell, collectingAndThen(counting(), Long::intValue)));
  }

  public static Set<Cell> nextGeneration(Set<Cell> liveCells) {
    var signals = generateSignalsForCells(liveCells);
    var counts = countSignals(signals);

    return counts.entrySet().stream()
      .filter(entry -> nextState(liveCells.contains(entry.getKey()) ? CellState.ALIVE : DEAD, entry.getValue()) == CellState.ALIVE)
      .map(Map.Entry::getKey)
      .collect(toSet());
  }
}
